from PIL import Image
import os

# File to store answered questions
ANSWERED_FILE = "answered_questions.txt"


def load_answered():
    if os.path.isfile(ANSWERED_FILE):
        with open(ANSWERED_FILE, "r") as f:
            return set(line.strip() for line in f if line.strip())
    return set()


def save_answered(question_text):
    with open(ANSWERED_FILE, "a") as f:
        f.write(question_text + "\n")


class QuizItem:
    def __init__(self, question_text, answer_text, question_image=None, answer_image=None):
        self.question_text = question_text
        self.answer_text = answer_text
        self.question_image = question_image
        self.answer_image = answer_image

    def display_image(self, image_path):
        if image_path and os.path.isfile(image_path):
            try:
                img = Image.open(image_path)
                img.show()
            except Exception as e:
                print(f"Error displaying image: {e}")
        else:
            print("No valid image to display.")

    def present_question(self):
        print("Question:")
        print(self.question_text)
        if self.question_image:
            print("Displaying question image...")
            self.display_image(self.question_image)

    def present_answer(self):
        print("Answer:")
        print(self.answer_text)
        if self.answer_image:
            print("Displaying answer image...")
            self.display_image(self.answer_image)


class QuizSection:
    def __init__(self, title, items):
        self.title = title
        self.items = items

    def run_section(self, answered_set):
        print("\n" + "=" * 50)
        print(f"Section: {self.title}")
        print("=" * 50)
        for item in self.items:
            if item.question_text in answered_set:
                continue

            print("\n" + "-" * 50)
            item.present_question()
            input("Press Enter to see the answer...")
            item.present_answer()

            user_input = input("Press Enter to continue to the next question, or type 's' to mark this question as answered (and skip in the future): ").strip().lower()
            if user_input == "s":
                save_answered(item.question_text)
                answered_set.add(item.question_text)
        print(f"End of section: {self.title}\n")


def run_quiz(sections):
    answered_set = load_answered()
    for section in sections:
        section.run_section(answered_set)
    print("Quiz finished!")


clocks = [
        QuizItem(
            question_text="What is a Logical Clock?",
            answer_text="An abstract mechanism to assign a value to events to provide a notion of ordering",
            ),
        QuizItem(
            question_text="What are the three key mechanisms for a logical clock",
            answer_text="1) How to incorporate an event into a clock \n 2) How to merge two clocks \n 3) How to compare two clocks?",
            ),
        QuizItem(
            question_text="What are the two most common logical clocks?",
            answer_text="Lamport Clock and Vector Clock",
            ),
        QuizItem(
            question_text="How does a Lamport Clock implement the three key mechanisms?",
            answer_text="1) Incorporation: increment the integer \n 2) Merging: Take the max of between two integers \n 3) Comparison: use < operator",
            ),
        QuizItem(
            question_text="How does a Vector Clock implement the three key mechanisms?",
            answer_text="1) Incorporation: increment the position in the vector of the host where the event occured \n 2) Merging: Pairwise max of the two vector clocks \n 3) Comparison: VC > VC' <=> (forall i: VC_i >= VC'_i) and (exists j: VC_j > VC'_j)",
            ),
        QuizItem(
            question_text="What type of ordering is a vector clock?",
            answer_text="Strict Partial Order",
            ),
        QuizItem(
            question_text="What type of ordering is a lamport clock (after breaking ties)?",
            answer_text="Total Order",
            ),
        ]
happens_before = [
        QuizItem(
            question_text="What is the formal definition of A happens before B (A -> B)",
            answer_text="A is a potential cause of B",
            ),
        QuizItem(
            question_text="What type of ordering is happens before?",
            answer_text="strict partial order",
            ),
        QuizItem(
            question_text="If neither A->B nor B->A holds, then what can be said about the relation of A and B?",
            answer_text="A || B",
            ),
        QuizItem(
            question_text="For the given vector clocks: \n A = [1 4 15] \n B = [3 1 2] \n C = [5 4 18] \n What can be said about the relationship between these events. Why?",
            answer_text="""\n1) C > A (A -> C) because for all i in N (where N is # of nodes), C[i] >= A[i] and there exists a j in N such that C[j] > A[j]
                           \n2) C > B (B -> C) because for all i in N (where N is # of nodes), C[i] >= B[i] and there exists a j in N such that C[j] > B[j]
                           \n3) A is concurrent with B (A || B) because there exists i in N such that A[i] > B[i] and there exists j in N such that B[j] > A[j].""",
            ),
        QuizItem(
            question_text="For two events A and B, we know A->B if... (What are the three conditions for a happens before relationship to be established?)",
            answer_text="""\n1) A and B are events on the same timeline, with B happening further down the timeline than A (Note: You can always establish a total order on the same timeline)
                           \n2) A is a message being sent, B is the same message being received
                           \n3) There exist an event C such that A->C, and C->B (Transitive property)
                            """
            ),
        QuizItem(
            question_text="What are the conditions for both the lamport and vector clock for happens before relationship to hold",
            answer_text="1) Lamport Clock: A -> B => LC(A) < LC (B) (and similarly) LC(A) > LC(B) => A not -> B \n 2) Vector Clock A -> B <=> VC(A) < VC(B)"
            ),
        QuizItem(
            question_text="If A || B and B || C, is A || C?",
            answer_text="No, concurrency is not transitive."
            ),
        ]
fifo = [
        QuizItem(
            question_text="What is the formal definition of the FIFO Property?",
            answer_text="If Host A sends messages X then Y to Host B => B delivers X, then Y."
            ),
        ]
causal_delivery = [
        QuizItem(
            question_text="What is the formal defintition of the Causal Property?",
            answer_text="If A and B are messages, and A->B => Any host delivering A and B delivers A first.",
            ),
        QuizItem(
            question_text="What type of property (safety vs liveness) is causal delivery?",
            answer_text="Safety Property (A violation has a finite witness)",
            ),
        # QuizItem(
        #     question_text="Draw a Lamport Diagram which shows a violation of causal delivery but no FIFO delivery.",
        #     answer_text="",
        #     answer_image="./causal_violation_fifo.png",
        #     )
        ]
orderings = [
        QuizItem(
            question_text="Draw a Lamport Diagram that witnesses a violation of causal ordering, but does not violate FIFO. Be able to explain why.",
            answer_text="",
            answer_image="./causal_violation_fifo.png",
            ),
         QuizItem(
            question_text="Draw a Lamport Diagram that witnesses a violation of both FIFO and Causal Ordering. Be able to explain why.",
            answer_text="",
            ),
         QuizItem(
            question_text="Draw a Lamport Diagram that witnesses a violation of total ordering but not causal ordering. Be able to describe the execution.",
            answer_text="",
            ),
         QuizItem(
            question_text="Draw a Lamport Diagram that witnesses a violation of causal ordering but not total ordering. Be able to describe the execution.",
            answer_text="",
            ),
        ]
snapshot = []
network_models = [
        QuizItem(
            question_text="What are the two main network models?",
            answer_text="1. Asynchronous \n 2.Synchronous",
            ),
        QuizItem(
            question_text="What is the definition of the synchronous model?",
            answer_text="There is an upper bound T on message delivery and processing time (this implies that time can be used as a failure detector)",
            ),
        QuizItem(
            question_text="What is the definition of the asynchronous model",
            answer_text="There does not exists an upper bound T (as defined in the synchronous model)",
            ),
]
fault_models = [
        QuizItem(
            question_text="What are the five fault models and what do they say?",
            answer_text="1. None: No Failure \n 2. Fail-Stop: Detectable failures \n 3. Crash: Failures but not entirely detectable \n 4. Omission: Messages may be lost \n 5. Byzantine: Anything can happen",
            ),
        ]
protocol_properties = [
        QuizItem(
            question_text="What is a Safety Property?",
            answer_text="A safety property is a proterty in which a bad things never happens in a system. If the property is violated, there exists a finite witness that demonstrates the failure. You can point to a specific point in time in which the property was not upheld.",
            ),
        QuizItem(
            question_text="What is a Liveness Property?",
            answer_text="Ensure that a good thing eventually happens in a system. If liveness is violated, there does not exist a finite witness of the failure, it requires infinite non-progress to be violated.",
            ),
        ]
reliable_delivery = [
        QuizItem(
            question_text="What is the formal definition of reliable delivery?",
            answer_text="If a host A sends a message M to host B, and not all messages are dropped, then B eventually delivers M",
            ),
        QuizItem(
            question_text="Is reliable delivery a safety or liveness property? Why?",
            answer_text="A liveness property since the message will *eventually* be delivered if not all messages are dropped.",
            ),
        QuizItem(
            question_text="What is an example way to implement a reliable delivery protocol?",
            answer_text="Repeatedly send a message (using some delay mechanism) until it has been acknowledged.",
            ),
        QuizItem(
            question_text="How do you maintain exactly-once delivery when employing a reliable delivery protocol?",
            answer_text="Ensure idempotency for received operations.",
            ),
        QuizItem(
            question_text="What is the definition of idempotency?",
            answer_text="Sending a message multiple times should have the same effect as if it was sent once.",
            ),
        QuizItem(
            question_text="How would you implement an exactly-once delivery protocol?",
            answer_text="Using a unique ID per message and only delivering messages with an unseen ID.",
            ),
        ]
reliable_broadcast = [
         QuizItem(
            question_text="Draw a Lamport Diagram that witnesses a violation of reliable broadcast. Be able to describe the execution.",
            answer_text="",
            ),
        ]
consistency = [
    QuizItem(
            question_text="What is are the three main types of consistency studied in class?",
            answer_text="1. Strong Consistency \n 2. Causal Consistency \n 3. Eventual Consistency",
            ),
    # QuizItem(
    #         question_text="What is are the three main types of consistency studied in class?",
    #         answer_text="1. Strong Consistency \n 2. Causal Consistency \n 3. Eventual Consistency",
    #         ),

    QuizItem(
            question_text="What type of property is Strong Consistency? Why?",
            answer_text="Strong Consistency is a safety property since a violation would require a fitness witness of the system behaving differently from a non-distributed system.",
            ),
    QuizItem(
            question_text="What type of property is Causal Consistency? Why?",
            answer_text="Similar to Strong Consistency, Causal Consistency is a safety property since a violation would require a fitness witness of the system behaving differently from a non-distributed system.",
            ),
    QuizItem(
            question_text="What type of property is Eventual Consistency? Why?",
            answer_text="Eventual Consistency is a Liveness Property (The Eventual keyword is a hint) since it only requires consistency eventually, meaning there does not exist a finite witness of a violatation of eventual consistency.",
            ),
    ]
write_conflicts = []
durability = []
primary_backup = []
chain_replication = []
state_machine_replication = []
quorum_replication = []
views = []
cap = []
two_generals = []
agreement = [

        QuizItem(
            question_text="What are the three properties of agreement?",
            answer_text="1) Agreement \n 2) Validity \n 3) Termination",
            ),
        QuizItem(
            question_text="What is the formal definition of agreement?",
            answer_text="If a node decides a value v, then all nodes decide v",
            ),
        QuizItem(
            question_text="What is the formal definition of validity?",
            answer_text="If a node decided v, then v was an input"
            ),
        QuizItem(
            question_text="What is the formal definition of termination?",
            answer_text="All nodes eventually decide"
            ),
        QuizItem(
            question_text="What is the FLP Impossibility statement?",
            answer_text="In an async model, it is impossible to achieve agreement, validility, and termination."
            ),
        QuizItem(
            question_text="What is Strong Agreement?",
            answer_text="A system will always agree and be valid, but may not terminate."
            ),
        QuizItem(
            question_text="What is Weak Agreement?",
            answer_text="A system will always terminate and be valid, but may not always agree."
            ),
        QuizItem(
            question_text="Is a agreement a saftey or liveness property? Why?",
            answer_text="Agreement is a safety property since there exists a finite witness of a violation of agreement",
            ),
        QuizItem(
            question_text="Is a validity a saftey or liveness property? Why?",
            answer_text="Validility is a safety property since there exists a finite witness of a violation of agreement",
            ),
        QuizItem(
            question_text="Is termination a safety or liveness property? Why?",
            answer_text="Termination is a liveness property since it stipulates that a system must eventually terminate which may take a long time, but there does not exist a finite witness of a violation of termination.",
            ),
        QuizItem(
            question_text="Draw a diagram that witnesses a violation of agreement. Be able to describe the execution.",
            answer_text="",
            ),
        QuizItem(
            question_text="Draw a diagram that witnesses a violation of validity. Be able to describe the execution.",
            answer_text="",
            ),
        ]
leader_election = [
        QuizItem(
            question_text="What agreement type does leader election satisfy.",
            answer_text="Leader election is a weak agreement protocol (meaning nodes may disagree on the leader)",
            ),
        QuizItem(
            question_text="How does Leader election determine that a node is down (failure detection)?",
            answer_text="Using heartbeats, nodes are pinging each other at certain time intervals (Usually constant). If a node doesn't reply in a certain time interval or after a number of heartbeats, that node is considered down",
            ),
        QuizItem(
            question_text="In Leader election, if a node determines the leader is down, what does it do?",
            answer_text="The node will initiate a leader election.",
            ),
        QuizItem(
            question_text="By what ordering do nodes usually determine the leader",
            answer_text="Nodes will usually choose that which has the lowest node id in the election process to be leader",
            ),
        QuizItem(
            question_text="Describe how a leader election works once a node initiates a leader election?",
            answer_text="The node that calls for the leader election will broadcast its node id, proposing to be leader. There are two cases for a node receiving the proposal: \n 1) The node id is higher, in which case it will ack the leader \n 2) The node id is lower, in which case it will broadcast itself to the other nodes as the proper leader. This occurs until a node think its been a leader for some fixed amount of time." 
            ),
        ]
atomic_commit = [
        QuizItem(
            question_text="Draw a Lamport Diagram that witnesses a violation of atomic commit. Be able to describe the execution.",
            answer_text="",
            ),
        QuizItem(
            question_text="Draw a Lamport Diagram that witnesses a violation of atomic commit. Be able to describe the execution.",
            answer_text="",
            ),

        ]
two_phase_commit = [
        QuizItem(
            question_text="Who are the two main actors in two-phase commit?",
            answer_text="Partcipants and the Coordinator (Only one)",
            ),
        QuizItem(
            question_text="What are the two phases of two-phase commit and what do they (generally) do?",
            answer_text="Phase 1: Preparation. In this stage, a transaction is sent by the coordinator to the partcipants and they attempt to prepare to perform the transaction (in an atomic way, eg. switching pointers). \n Phase 2: Action. In this stage, a transaction is either commited and the message to commit is sent to the partcipants or the message is aborted and an abort message is sent to the partcipants.",
            ),
        QuizItem(
            question_text="What if any partcipants replies with abort in the preparation phase?",
            answer_text="All participants will abort. (Either via a message from the coordinator or from a Collaborative Termination Protocol)",
            ),
        QuizItem(
            question_text="How would a coordinator deal with a non-responsive participant (or the messages get lost)?",
            answer_text="The coordinator has a fixed timeout, after which it will abort the transaction.",
            ),
        QuizItem(
            question_text="Draw a Lamport Diagram that witnesses a violation of atomic commit that uses a two phase commit process. Be able to describe the execution.",
            answer_text="",
            ),
        QuizItem(
            question_text="Draw a Lamport Diagram in which two-phase commit will not terminate.",
            answer_text="",
            ),
        QuizItem(
            question_text="Draw a Lamport Diagram in which two-phase commit performs a commit with three participant nodes.",
            answer_text="",
            ),
        ]
consensus = []
paxos = []
multipaxos = []
sharding = []
consistent_hasing = []

sections = [
    QuizSection("Clocks", clocks),
    QuizSection("Happens Before", happens_before),
    QuizSection("FIFO", fifo),
    QuizSection("Causal Delivery", causal_delivery),
    QuizSection("Orderings", orderings),
    QuizSection("Snapshot", snapshot),
    QuizSection("Network Models", network_models),
    QuizSection("Fault Models", fault_models),
    QuizSection("Reliable Delivery", reliable_delivery),
    QuizSection("Protocol Properties", protocol_properties),
    QuizSection("Reliable Broadcast", reliable_broadcast),
    QuizSection("Consistency", consistency),
    QuizSection("Durability", durability),
    QuizSection("Primary Backup", primary_backup),
    QuizSection("Chain Replication", chain_replication),
    QuizSection("State Machine Replication", state_machine_replication),
    QuizSection("Quorum Replication", quorum_replication),
    QuizSection("Views", views),
    QuizSection("CAP", cap),
    QuizSection("Two Generals", two_generals),
    QuizSection("Agreement", agreement),
    QuizSection("Leader Election", leader_election),
    QuizSection("Atomic Commit", atomic_commit),
    QuizSection("Two Phase Commit", two_phase_commit),
    QuizSection("Consensus", consensus),
    QuizSection("Paxos", paxos),
    QuizSection("MultiPaxos", multipaxos),
    QuizSection("Sharding", sharding),
    QuizSection("Consistent Hashing", consistent_hasing),
]


if __name__ == "__main__":
    run_quiz(sections)
